@@ -1,9 +1,5 @@
 # frozen_string_literal: true
 
-require 'rack/request'
-require 'rack/utils'
-require 'rack/body_proxy'
-require 'rack/media_type'
 require 'time'
 
 module Rack
@@ -19,34 +15,51 @@ module Rack
   # +write+ are synchronous with the Rack response.
   #
   # Your application's +call+ should end returning Response#finish.
-
   class Response
-    attr_accessor :length, :status, :body
-    attr_reader :header
-    alias headers header
+    def self.[](status, headers, body)
+      self.new(body, status, headers)
+    end
 
     CHUNKED = 'chunked'
     STATUS_WITH_NO_ENTITY_BODY = Utils::STATUS_WITH_NO_ENTITY_BODY
 
-    def initialize(body = nil, status = 200, header = {})
+    attr_accessor :length, :status, :body
+    attr_reader :headers
+
+    # @deprecated Use {#headers} instead.
+    alias header headers
+
+    # Initialize the response object with the specified body, status
+    # and headers.
+    #
+    # @param body [nil, #each, #to_str] the response body.
+    # @param status [Integer] the integer status as defined by the
+    # HTTP protocol RFCs.
+    # @param headers [#each] a list of key-value header pairs which
+    # conform to the HTTP protocol RFCs.
+    #
+    # Providing a body which responds to #to_str is legacy behaviour.
+    def initialize(body = nil, status = 200, headers = {})
       @status = status.to_i
-      @header = Utils::HeaderHash.new(header)
+      @headers = Utils::HeaderHash[headers]
 
       @writer = self.method(:append)
 
       @block = nil
-      @length = 0
 
       # Keep track of whether we have expanded the user supplied body.
       if body.nil?
         @body = []
         @buffered = true
+        @length = 0
       elsif body.respond_to?(:to_str)
         @body = [body]
         @buffered = true
+        @length = body.to_str.bytesize
       else
         @body = body
         @buffered = false
+        @length = 0
       end
 
       yield self if block_given?
@@ -61,18 +74,21 @@ module Rack
       CHUNKED == get_header(TRANSFER_ENCODING)
     end
 
+    # Generate a response array consistent with the requirements of the SPEC.
+    # @return [Array] a 3-tuple suitable of `[status, headers, body]`
+    # which is suitable to be returned from the middleware `#call(env)` method.
     def finish(&block)
       if STATUS_WITH_NO_ENTITY_BODY[status.to_i]
         delete_header CONTENT_TYPE
         delete_header CONTENT_LENGTH
         close
-        [status.to_i, header, []]
+        return [@status, @headers, []]
       else
         if block_given?
           @block = block
-          [status.to_i, header, self]
+          return [@status, @headers, self]
         else
-          [status.to_i, header, @body]
+          return [@status, @headers, @body]
         end
       end
     end
@@ -152,7 +168,7 @@ module Rack
       #   assert_equal 'Accept-Encoding,Cookie', response.get_header('Vary')
       #
       # http://www.w3.org/Protocols/rfc2616/rfc2616-sec4.html#sec4.2
-      def add_header key, v
+      def add_header(key, v)
         if v.nil?
           get_header key
         elsif has_header? key
@@ -162,10 +178,16 @@ module Rack
         end
       end
 
+      # Get the content type of the response.
       def content_type
         get_header CONTENT_TYPE
       end
 
+      # Set the content type of the response.
+      def content_type=(content_type)
+        set_header CONTENT_TYPE, content_type
+      end
+
       def media_type
         MediaType.type(content_type)
       end
@@ -200,7 +222,7 @@ module Rack
         get_header SET_COOKIE
       end
 
-      def set_cookie_header= v
+      def set_cookie_header=(v)
         set_header SET_COOKIE, v
       end
 
@@ -208,15 +230,31 @@ module Rack
         get_header CACHE_CONTROL
       end
 
-      def cache_control= v
+      def cache_control=(v)
         set_header CACHE_CONTROL, v
       end
 
+      # Specifies that the content shouldn't be cached. Overrides `cache!` if already called.
+      def do_not_cache!
+        set_header CACHE_CONTROL, "no-cache, must-revalidate"
+        set_header EXPIRES, Time.now.httpdate
+      end
+
+      # Specify that the content should be cached.
+      # @param duration [Integer] The number of seconds until the cache expires.
+      # @option directive [String] The cache control directive, one of "public", "private", "no-cache" or "no-store".
+      def cache!(duration = 3600, directive: "public")
+        unless headers[CACHE_CONTROL] =~ /no-cache/
+          set_header CACHE_CONTROL, "#{directive}, max-age=#{duration}"
+          set_header EXPIRES, (Time.now + duration).httpdate
+        end
+      end
+
       def etag
         get_header ETAG
       end
 
-      def etag= v
+      def etag=(v)
         set_header ETAG, v
       end
 
@@ -228,6 +266,9 @@ module Rack
         if @body.is_a?(Array)
           # The user supplied body was an array:
           @body = @body.compact
+          @body.each do |part|
+            @length += part.to_s.bytesize
+          end
         else
           # Turn the user supplied body into a buffered array:
           body = @body
@@ -236,6 +277,8 @@ module Rack
           body.each do |part|
             @writer.call(part.to_s)
           end
+
+          body.close if body.respond_to?(:close)
         end
 
         @buffered = true
@@ -261,7 +304,7 @@ module Rack
       attr_reader :headers
       attr_accessor :status
 
-      def initialize status, headers
+      def initialize(status, headers)
         @status = status
         @headers = headers
       end
