@@ -1,11 +1,6 @@
 # frozen_string_literal: true
 
-require 'minitest/global_expectations/autorun'
-require 'rack/builder'
-require 'rack/lint'
-require 'rack/mock'
-require 'rack/show_exceptions'
-require 'rack/urlmap'
+require_relative 'helper'
 
 class NothingMiddleware
   def initialize(app, **)
@@ -43,6 +38,19 @@ describe Rack::Builder do
     Rack::MockRequest.new(app).get("/sub").body.to_s.must_equal 'sub'
   end
 
+  it "supports use when mapping" do
+    app = builder_to_app do
+      map '/sub' do
+        use Rack::ContentLength
+        run lambda { |inner_env| [200, { "Content-Type" => "text/plain" }, ['sub']] }
+      end
+      use Rack::ContentLength
+      run lambda { |inner_env| [200, { "Content-Type" => "text/plain" }, ['root']] }
+    end
+    Rack::MockRequest.new(app).get("/").headers['Content-Length'].must_equal '4'
+    Rack::MockRequest.new(app).get("/sub").headers['Content-Length'].must_equal '3'
+  end
+
   it "doesn't dupe env even when mapping" do
     app = builder_to_app do
       use NothingMiddleware, noop: :noop
@@ -253,8 +261,23 @@ describe Rack::Builder do
     end
 
     it "strips leading unicode byte order mark when present" do
-      app, _ = Rack::Builder.parse_file config_file('bom.ru')
-      Rack::MockRequest.new(app).get("/").body.to_s.must_equal 'OK'
+      enc = Encoding.default_external
+      begin
+        Encoding.default_external = 'UTF-8'
+        app, _ = Rack::Builder.parse_file config_file('bom.ru')
+        Rack::MockRequest.new(app).get("/").body.to_s.must_equal 'OK'
+      ensure
+        Encoding.default_external = enc
+      end
+    end
+
+    it "respects the frozen_string_literal magic comment" do
+      app, _ = Rack::Builder.parse_file(config_file('frozen.ru'))
+      response = Rack::MockRequest.new(app).get('/')
+      response.body.must_equal 'frozen'
+      body = response.instance_variable_get(:@body)
+      body.must_equal(['frozen'])
+      body[0].frozen?.must_equal true
     end
   end
 
