@@ -1,7 +1,6 @@
 # frozen_string_literal: true
 
-require 'minitest/global_expectations/autorun'
-require 'rack/handler'
+require_relative 'helper'
 
 class Rack::Handler::Lobster; end
 class RockLobster; end
