@@ -1,10 +1,7 @@
 # frozen_string_literal: true
 
-require 'minitest/global_expectations/autorun'
-require 'stringio'
+require_relative 'helper'
 require 'tempfile'
-require 'rack/lint'
-require 'rack/mock'
 
 describe Rack::Lint do
   def env(*args)
@@ -26,6 +23,10 @@ describe Rack::Lint do
     lambda { Rack::Lint.new(nil).call 5 }.must_raise(Rack::Lint::LintError).
       message.must_match(/not a Hash/)
 
+    lambda { Rack::Lint.new(nil).call({}.freeze) }.must_raise(Rack::Lint::LintError).
+      message.must_match(/env should not be frozen, but is/)
+
+
     lambda {
       e = env
       e.delete("REQUEST_METHOD")
@@ -72,10 +73,66 @@ describe Rack::Lint do
       message.must_equal "session [] must respond to store and []="
 
     lambda {
-      Rack::Lint.new(nil).call(env("rack.logger" => []))
+      Rack::Lint.new(nil).call(env("rack.session" => {}.freeze))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal "session {} must respond to to_hash and return unfrozen Hash instance"
+
+    obj = {}
+    obj.singleton_class.send(:undef_method, :to_hash)
+    lambda {
+      Rack::Lint.new(nil).call(env("rack.session" => obj))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal "session {} must respond to to_hash and return unfrozen Hash instance"
+
+    obj.singleton_class.send(:undef_method, :clear)
+    lambda {
+      Rack::Lint.new(nil).call(env("rack.session" => obj))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal "session {} must respond to clear"
+
+    obj.singleton_class.send(:undef_method, :delete)
+    lambda {
+      Rack::Lint.new(nil).call(env("rack.session" => obj))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal "session {} must respond to delete"
+
+    obj.singleton_class.send(:undef_method, :fetch)
+    lambda {
+      Rack::Lint.new(nil).call(env("rack.session" => obj))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal "session {} must respond to fetch and []"
+
+    obj = Object.new
+    def obj.inspect; '[]' end
+    lambda {
+      Rack::Lint.new(nil).call(env("rack.logger" => obj))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "logger [] must respond to info"
 
+    def obj.info(*) end
+    lambda {
+      Rack::Lint.new(nil).call(env("rack.logger" => obj))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal "logger [] must respond to debug"
+
+    def obj.debug(*) end
+    lambda {
+      Rack::Lint.new(nil).call(env("rack.logger" => obj))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal "logger [] must respond to warn"
+
+    def obj.warn(*) end
+    lambda {
+      Rack::Lint.new(nil).call(env("rack.logger" => obj))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal "logger [] must respond to error"
+
+    def obj.error(*) end
+    lambda {
+      Rack::Lint.new(nil).call(env("rack.logger" => obj))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal "logger [] must respond to fatal"
+
     lambda {
       Rack::Lint.new(nil).call(env("rack.multipart.buffer_size" => 0))
     }.must_raise(Rack::Lint::LintError).
@@ -93,11 +150,24 @@ describe Rack::Lint do
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "rack.multipart.tempfile_factory return value must respond to #<<"
 
+    lambda {
+      Rack::Lint.new(lambda { |env|
+        env['rack.multipart.tempfile_factory'].call("testfile", "text/plain")
+        []
+      }).call(env("rack.multipart.tempfile_factory" => lambda { |filename, content_type| String.new }))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal "response array has 0 elements instead of 3"
+
     lambda {
       Rack::Lint.new(nil).call(env("REQUEST_METHOD" => "FUCKUP?"))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/REQUEST_METHOD/)
 
+    lambda {
+      Rack::Lint.new(nil).call(env("REQUEST_METHOD" => "OOPS?\b!"))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_match(/OOPS\?\\/)
+
     lambda {
       Rack::Lint.new(nil).call(env("SCRIPT_NAME" => "howdy"))
     }.must_raise(Rack::Lint::LintError).
@@ -113,6 +183,20 @@ describe Rack::Lint do
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/Invalid CONTENT_LENGTH/)
 
+    lambda {
+      Rack::Lint.new(nil).call(env("QUERY_STRING" => nil))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_include('env variable QUERY_STRING has non-string value nil')
+
+    lambda {
+      Rack::Lint.new(nil).call(env("QUERY_STRING" => "\u1234"))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_include('env variable QUERY_STRING has value containing non-ASCII characters and has non-ASCII-8BIT encoding')
+
+    Rack::Lint.new(lambda { |env|
+                     [200, {}, []]
+                   }).call(env("QUERY_STRING" => "\u1234".b)).first.must_equal 200
+
     lambda {
       e = env
       e.delete("PATH_INFO")
@@ -158,11 +242,29 @@ describe Rack::Lint do
 
   it "notice error errors" do
     lambda {
-      Rack::Lint.new(nil).call(env("rack.errors" => ""))
+      io = StringIO.new
+      io.binmode
+      Rack::Lint.new(nil).call(env("rack.errors" => "", "rack.input" => io))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/does not respond to #puts/)
   end
 
+  it "notice response errors" do
+    lambda {
+      Rack::Lint.new(lambda { |env|
+                       ""
+                     }).call(env({}))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_include('response is not an Array, but String')
+
+    lambda {
+      Rack::Lint.new(lambda { |env|
+                       [nil, nil, nil, nil]
+                     }).call(env({}))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_include('response array has 4 elements instead of 3')
+  end
+
   it "notice status errors" do
     lambda {
       Rack::Lint.new(lambda { |env|
@@ -181,9 +283,12 @@ describe Rack::Lint do
 
   it "notice header errors" do
     lambda {
+      io = StringIO.new('a')
+      io.binmode
       Rack::Lint.new(lambda { |env|
+                       env['rack.input'].each{ |x| }
                        [200, Object.new, []]
-                     }).call(env({}))
+                     }).call(env({ "rack.input" => io }))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "headers object should respond to #each, but doesn't (got Object as headers)"
 
@@ -320,6 +425,7 @@ describe Rack::Lint do
 
     lambda {
       Rack::Lint.new(lambda { |env|
+                       env["rack.input"].gets
                        env["rack.input"].read(1, 2, 3)
                        [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
                      }).call(env({}))
@@ -503,6 +609,28 @@ describe Rack::Lint do
     assert_lint nil, ''.dup
     assert_lint 1, ''.dup
   end
+
+  it "notice hijack errors" do
+    lambda {
+      Rack::Lint.new(lambda { |env|
+                       env['rack.hijack'].call
+                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                     }).call(env({ 'rack.hijack?' => true, 'rack.hijack' => lambda { Object.new } }))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_match(/rack.hijack_io must respond to read/)
+
+      Rack::Lint.new(lambda { |env|
+                       env['rack.hijack'].call
+                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                     }).call(env({ 'rack.hijack?' => true, 'rack.hijack' => lambda { StringIO.new }, 'rack.hijack_io' => StringIO.new })).
+        first.must_equal 201
+
+      Rack::Lint.new(lambda { |env|
+                       env['rack.hijack?'] = true
+                       [201, { "Content-type" => "text/plain", "Content-length" => "0", 'rack.hijack' => lambda {|io| io }, 'rack.hijack_io' => StringIO.new }, []]
+                     }).call(env({}))[1]['rack.hijack'].call(StringIO.new).read.must_equal ''
+  end
+
 end
 
 describe "Rack::Lint::InputWrapper" do
