@@ -121,6 +121,11 @@ class RackRequestTest < Minitest::Spec
     req.host.must_equal "123foo.example.com"
     req.hostname.must_equal "123foo.example.com"
 
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "some_service:3001")
+    req.host.must_equal "some_service"
+    req.hostname.must_equal "some_service"
+
     req = make_request \
       Rack::MockRequest.env_for("/", "SERVER_NAME" => "example.org", "SERVER_PORT" => "9292")
     req.host.must_equal "example.org"
@@ -156,6 +161,10 @@ class RackRequestTest < Minitest::Spec
       Rack::MockRequest.env_for("/", "HTTP_HOST" => "www2.example.org:81")
     req.port.must_equal 81
 
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "some_service:3001")
+    req.port.must_equal 3001
+
     req = make_request \
       Rack::MockRequest.env_for("/", "SERVER_NAME" => "example.org", "SERVER_PORT" => "9292")
     req.port.must_equal 9292
