@@ -1,24 +1,22 @@
 # frozen_string_literal: true
 
-require 'minitest/global_expectations/autorun'
-require 'rack/utils'
-require 'rack/mock'
+require_relative 'helper'
 require 'timeout'
 
 describe Rack::Utils do
 
-  def assert_sets exp, act
+  def assert_sets(exp, act)
     exp = Set.new exp.split '&'
     act = Set.new act.split '&'
 
     assert_equal exp, act
   end
 
-  def assert_query exp, act
+  def assert_query(exp, act)
     assert_sets exp, Rack::Utils.build_query(act)
   end
 
-  def assert_nested_query exp, act
+  def assert_nested_query(exp, act)
     assert_sets exp, Rack::Utils.build_nested_query(act)
   end
 
@@ -36,7 +34,7 @@ describe Rack::Utils do
 
   it "round trip binary data" do
     r = [218, 0].pack 'CC'
-      z = Rack::Utils.unescape(Rack::Utils.escape(r), Encoding::BINARY)
+    z = Rack::Utils.unescape(Rack::Utils.escape(r), Encoding::BINARY)
     r.must_equal z
   end
 
@@ -106,6 +104,12 @@ describe Rack::Utils do
     Rack::Utils.parse_query(",foo=bar;,", ";,").must_equal "foo" => "bar"
   end
 
+  it "parse query strings correctly using arrays" do
+    Rack::Utils.parse_query("a[]=1").must_equal "a[]" => "1"
+    Rack::Utils.parse_query("a[]=1&a[]=2").must_equal "a[]" => ["1", "2"]
+    Rack::Utils.parse_query("a[]=1&a[]=2&a[]=3").must_equal "a[]" => ["1", "2", "3"]
+  end
+
   it "not create infinite loops with cycle structures" do
     ex = { "foo" => nil }
     ex["foo"] = ex
@@ -124,11 +128,17 @@ describe Rack::Utils do
 
     lambda {
       Rack::Utils.parse_nested_query("foo#{"[a]" * len}=bar")
-    }.must_raise(RangeError)
+    }.must_raise(Rack::QueryParser::ParamsTooDeepError)
 
     Rack::Utils.parse_nested_query("foo#{"[a]" * (len - 1)}=bar")
   end
 
+  # ParamsTooDeepError was introduced in the middle of 2.2 releases
+  # and this test is here to ensure backwards compatibility
+  it "ParamsTooDeepError is inherited from originally used RangeError" do
+    (Rack::QueryParser::ParamsTooDeepError < RangeError).must_equal(true)
+  end
+
   it "parse nested query strings correctly" do
     Rack::Utils.parse_nested_query("foo").
       must_equal "foo" => nil
@@ -254,8 +264,12 @@ describe Rack::Utils do
         end
       end
       Rack::Utils.default_query_parser = Rack::QueryParser.new(param_parser_class, 65536, 100)
-      Rack::Utils.parse_query(",foo=bar;,", ";,")[:foo].must_equal "bar"
-      Rack::Utils.parse_nested_query("x[y][][z]=1&x[y][][w]=2")[:x][:y][0][:z].must_equal "1"
+      h1 = Rack::Utils.parse_query(",foo=bar;,", ";,")
+      h1[:foo].must_equal "bar"
+      h2 = Rack::Utils.parse_nested_query("x[y][][z]=1&x[y][][w]=2")
+      h2[:x][:y][0][:z].must_equal "1"
+      h3 = Rack::Utils.parse_nested_query("")
+      h3.merge(h1)[:foo].must_equal "bar"
     ensure
       Rack::Utils.default_query_parser = default_parser
     end
@@ -436,6 +450,7 @@ describe Rack::Utils do
 
     helper.call(%w(compress gzip identity), [["compress", 1.0], ["gzip", 1.0]]).must_equal "compress"
     helper.call(%w(compress gzip identity), [["compress", 0.5], ["gzip", 1.0]]).must_equal "gzip"
+    helper.call(%w(compress gzip identity), [["gzip", 1.0], ["compress", 1.0]]).must_equal "compress"
 
     helper.call(%w(foo bar identity), []).must_equal "identity"
     helper.call(%w(foo bar identity), [["*", 1.0]]).must_equal "foo"
@@ -510,6 +525,9 @@ describe Rack::Utils, "cookies" do
     env = Rack::MockRequest.env_for("", "HTTP_COOKIE" => "foo=bar;quux=h&m")
     Rack::Utils.parse_cookies(env).must_equal({ "foo" => "bar", "quux" => "h&m" })
 
+    env = Rack::MockRequest.env_for("", "HTTP_COOKIE" => "foo=bar; quux=h&m")
+    Rack::Utils.parse_cookies(env).must_equal({ "foo" => "bar", "quux" => "h&m" })
+
     env = Rack::MockRequest.env_for("", "HTTP_COOKIE" => "foo=bar").freeze
     Rack::Utils.parse_cookies(env).must_equal({ "foo" => "bar" })
 
@@ -545,6 +563,30 @@ describe Rack::Utils, "cookies" do
       Rack::Utils.add_cookie_to_header(Object.new, 'name', 'value')
     }.must_raise ArgumentError
   end
+
+  it "sets and deletes cookies in header hash" do
+    header = { 'Set-Cookie' => '' }
+    Rack::Utils.set_cookie_header!(header, 'name', 'value').must_be_nil
+    header['Set-Cookie'].must_equal 'name=value'
+    Rack::Utils.set_cookie_header!(header, 'name2', 'value2').must_be_nil
+    header['Set-Cookie'].must_equal "name=value\nname2=value2"
+    Rack::Utils.set_cookie_header!(header, 'name2', 'value3').must_be_nil
+    header['Set-Cookie'].must_equal "name=value\nname2=value2\nname2=value3"
+
+    Rack::Utils.delete_cookie_header!(header, 'name2').must_be_nil
+    header['Set-Cookie'].must_equal "name=value\nname2=; max-age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT"
+    Rack::Utils.delete_cookie_header!(header, 'name').must_be_nil
+    header['Set-Cookie'].must_equal "name2=; max-age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT\nname=; max-age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT"
+
+    header = { 'Set-Cookie' => nil }
+    Rack::Utils.delete_cookie_header!(header, 'name').must_be_nil
+    header['Set-Cookie'].must_equal "name=; max-age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT"
+
+    header = { 'Set-Cookie' => [] }
+    Rack::Utils.delete_cookie_header!(header, 'name').must_be_nil
+    header['Set-Cookie'].must_equal "name=; max-age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT"
+  end
+
 end
 
 describe Rack::Utils, "byte_range" do
@@ -670,7 +712,7 @@ describe Rack::Utils::HeaderHash do
     h.delete("Foo").must_equal "bar"
   end
 
-  it "return nil when #delete is called on a non-existant key" do
+  it "return nil when #delete is called on a non-existent key" do
     h = Rack::Utils::HeaderHash.new("foo" => "bar")
     h.delete("Hello").must_be_nil
   end
@@ -696,14 +738,43 @@ describe Rack::Utils::HeaderHash do
     h['foo'].must_be_nil
     h.wont_include 'foo'
   end
+
+  it "uses memoized header hash" do
+    env = {}
+    headers = Rack::Utils::HeaderHash.new({ 'content-type' => "text/plain", "content-length" => "3" })
+
+    app = lambda do |env|
+      [200, headers, []]
+    end
+
+    app = Rack::ContentLength.new(app)
+
+    response = app.call(env)
+    assert_same response[1], headers
+  end
+
+  it "duplicates header hash" do
+    env = {}
+    headers = Rack::Utils::HeaderHash.new({ 'content-type' => "text/plain", "content-length" => "3" })
+    headers.freeze
+
+    app = lambda do |env|
+      [200, headers, []]
+    end
+
+    app = Rack::ContentLength.new(app)
+
+    response = app.call(env)
+    refute_same response[1], headers
+  end
 end
 
 describe Rack::Utils::Context do
   class ContextTest
     attr_reader :app
-    def initialize app; @app = app; end
-    def call env; context env; end
-    def context env, app = @app; app.call(env); end
+    def initialize(app); @app = app; end
+    def call(env); context env; end
+    def context(env, app = @app); app.call(env); end
   end
   test_target1 = proc{|e| e.to_s + ' world' }
   test_target2 = proc{|e| e.to_i + 2 }
@@ -744,6 +815,8 @@ describe Rack::Utils::Context do
     r2.must_equal 4
     r3 = c3.call(:misc_symbol)
     r3.must_be_nil
+    r3 = c2.context(:misc_symbol, test_target3)
+    r3.must_be_nil
     r4 = Rack::MockRequest.new(a4).get('/')
     r4.status.must_equal 200
     r5 = Rack::MockRequest.new(a5).get('/')
